// __BEGIN_LICENSE__
// Copyright (C) 2006-2011 United States Government as represented by
// the Administrator of the National Aeronautics and Space Administration.
// All Rights Reserved.
// __END_LICENSE__


// TestPinholeModel.h
#include <gtest/gtest.h>

#include <boost/random.hpp>

#include <vw/Math/Vector.h>
#include <vw/Math/EulerAngles.h>
#include <vw/Math/LinearAlgebra.h>

#include <vw/Camera/PinholeModel.h>
#include <vw/Camera/CameraTransform.h>
#include <vw/Camera/CameraGeometry.h>

#include <test/Helpers.h>

using namespace vw;
using namespace vw::camera;
using namespace vw::test;

TEST( PinholeModel, StandardConstruct ) {
  Matrix<double,3,3> pose;
  pose.set_identity();

  // Create an imaginary 1000x1000 pixel imager
  PinholeModel pinhole( Vector3(0,0,0), // camera center
                        pose,           // camera pose
                        500,500,        // fx, fy
                        500,500,        // cx, cy
                        NullLensDistortion());

  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(0,0,10)),
                     Vector2(500,500), 1e-6);
  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(-10,0,10)),
                     Vector2(0,500), 1e-6);
  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(10,0,10)),
                     Vector2(1000,500), 1e-6);
  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(0,-10,10)),
                     Vector2(500,1000), 1e-6);
  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(0,10,10)),
                     Vector2(500,0), 1e-6);
  EXPECT_STREQ( "Pinhole", pinhole.type().c_str() );
}

TEST( PinholeModel, CoordinateFrame ) {
  Matrix<double,3,3> pose;
  pose.set_identity();

  // Create an imaginary 1000x1000 pixel imager, where the camera
  // coordinate system is mapped as follows:
  //
  // +u : along the camera +Y axis
  // +v : along the camera +X axis
  // +w : along the camera -Z axis
  PinholeModel pinhole( Vector3(0,0,0), // camera center
                        pose,           // camera pose
                        500,500,        // fx, fy
                        500,500,
                        Vector3(0, 1, 0),
                        Vector3(1, 0, 0),
                        Vector3(0, 0, -1),
                        NullLensDistortion());       // cx, cy

  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(0,0,-10)),
                     Vector2(500,500), 1e-5);
  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(-10,0,-10)),
                     Vector2(500,0), 1e-5);
  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(10,0,-10)),
                     Vector2(500,1000), 1e-5);
  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(0,-10,-10)),
                     Vector2(0,500), 1e-5);
  EXPECT_VECTOR_NEAR(pinhole.point_to_pixel(Vector3(0,10,-10)),
                     Vector2(1000,500), 1e-5);
}

TEST( PinholeModel, TsaiDistortion ) {
  // Create an imaginary 1000x1000 pixel imager
  PinholeModel pinhole( Vector3(0,0,0),                 // camera center
                        math::identity_matrix<3>(),     // camera pose
                        500,500,                        // fx, fy
                        500,500,                        // cx, cy
                        TsaiLensDistortion(Vector4(-0.2805362343788147,
                                                   0.1062035113573074,
                                                   -0.0001422458299202845,
                                                   0.00116333004552871)));
  const LensDistortion* distortion = pinhole.lens_distortion();

#if defined(VW_HAVE_PKG_LAPACK) && VW_HAVE_PKG_LAPACK==1
  Vector2 distorted_pix = distortion->distorted_coordinates(pinhole, Vector2(200,200));
  Vector2 undistorted_pix = distortion->undistorted_coordinates(pinhole, distorted_pix);

  EXPECT_VECTOR_NEAR( distorted_pix,
                      Vector2(244.865,244.395),
                      1e-1 );
  EXPECT_VECTOR_NEAR( undistorted_pix,
                      Vector2(200,200),
                      1e-1 );
#endif
}

TEST( PinholeModel, ScalePinhole ) {
  Matrix<double,3,3> rot = vw::math::euler_to_quaternion(1.15, 0.0, -1.57, "xyz").rotation_matrix();
  PinholeModel pinhole4(Vector3(-0.329, 0.065, -0.82),
                        rot,
                        605.320556640625,
                        606.3638305664062,
                        518.89208984375,
                        387.5555114746094,
                        TsaiLensDistortion(Vector4(-0.2796604335308075,
                                                   0.1031486615538597,
                                                   -0.0007824968779459596,
                                                   0.0009675505571067333)));
  PinholeModel scaled = scale_camera(pinhole4, .1);

  Vector3 point = Vector3(2,-1,1) +
    5*pinhole4.pixel_to_vector(Vector2(500,500)) +
    pinhole4.camera_center();

  Vector2 o_return = pinhole4.point_to_pixel(point);
  Vector2 s_return = scaled.point_to_pixel(point);

  EXPECT_NEAR(o_return[0],s_return[0]*10,5); // Lens distortion doesn't
  EXPECT_NEAR(o_return[1],s_return[1]*10,5); // seem to be too accurate
}

TEST( PinholeModel, ProjectiveMatrix ) {
  // First set control camera
  Matrix<double,3,3> pose = math::euler_to_rotation_matrix(1.3,2.0,-.7,"xyz");

  // Create an imaginary 1000x1000 pixel imager
  PinholeModel control_pinhole( Vector3(0,4,-10),
                                pose, 600, 700,
                                500, 500,
                                NullLensDistortion() );

  // Make solve control camera (w/ random input for now)
  pose.set_identity();
  PinholeModel solved_pinhole( Vector3(-5,-2,5),
                               pose, 833, 544,
                               400, 700,
                               NullLensDistortion() );

  // Create Measurements used to solve for camera matrix
  std::vector<Vector<double> > world_m, image_m;
  boost::minstd_rand random_gen(42u);
  boost::normal_distribution<double> normal(0,20);
  boost::variate_generator<boost::minstd_rand&,
    boost::normal_distribution<double> > generator( random_gen, normal );
  for ( uint8 i = 0; i < 6; i++ ) {
    Vector3 point( generator(), generator(), generator()+60.0 );
    world_m.push_back( Vector4(point[0],point[1],point[2],1) );
    Vector2 pixel = control_pinhole.point_to_pixel(point);
    image_m.push_back( Vector3(pixel[0],pixel[1],1) );
  }

  // Building Camera Matrix
  CameraMatrixFittingFunctor fitfunc;
  Matrix<double> P = fitfunc(world_m,image_m);

  solved_pinhole.set_camera_matrix( P );

  // Compare camera matrices
  {
    Vector3 solved = solved_pinhole.camera_center( Vector2() );
    Vector3 control = control_pinhole.camera_center( Vector2() );
    EXPECT_VECTOR_NEAR( solved,
                        control, 1e-8 );
  }
  {
    Quaternion<double> solved = solved_pinhole.camera_pose( Vector2() );
    Quaternion<double> control = control_pinhole.camera_pose( Vector2() );
    for ( uint32 i = 0; i < 4; i ++ )
      EXPECT_NEAR( solved[i], control[i], 1e-8 );
  }
  {
    Vector2 s_focal = solved_pinhole.focal_length();
    Vector2 c_focal = control_pinhole.focal_length();
    EXPECT_NEAR( s_focal[0]/s_focal[1],c_focal[0]/c_focal[1], 1e-2 );
    EXPECT_VECTOR_NEAR( elem_quot(solved_pinhole.point_offset(),s_focal),
                        elem_quot(control_pinhole.point_offset(),c_focal), 1e-2 );
  }
}

class PinholeTest : public ::testing::Test {
protected:
  PinholeTest() {}

  virtual void SetUp() {
    expect_rot = vw::math::euler_to_rotation_matrix(1.15, 0.0, -1.57, "xyz");
    pinhole = PinholeModel(Vector3(-0.329, 0.065, -0.82),
                           expect_rot,
                           605.320556640625,
                           606.3638305664062,
                           518.89208984375,
                           387.5555114746094);
  }

  void projection_test(double tolerance=1e-6) {
    Vector2 image_size = pinhole.point_offset();
    image_size *= 2;
    for ( unsigned x = 10; x < image_size.x(); x+=80 ) {
      for ( unsigned y = 10; y < image_size.y(); y+=80 ) {
          EXPECT_VECTOR_NEAR( Vector2(x,y),
                              pinhole.point_to_pixel(pinhole.pixel_to_vector(Vector2(x,y)) +
                                                     pinhole.camera_center(Vector2(x,y))),
                            tolerance );
      }
    }
  }

  void readback_test(std::string const& file) {
#if defined(VW_HAVE_PKG_PROTOBUF) && VW_HAVE_PKG_PROTOBUF==1
    pinhole.write( file );
    PinholeModel read_back;
    read_back.read( file );
    EXPECT_VECTOR_DOUBLE_EQ( read_back.camera_center(),
                             pinhole.camera_center() );
    Matrix<double> read_back_R =
      read_back.camera_pose().rotation_matrix();
    EXPECT_MATRIX_NEAR( read_back_R, expect_rot,1e-8 );
    EXPECT_VECTOR_DOUBLE_EQ( read_back.coordinate_frame_u_direction(),
                             pinhole.coordinate_frame_u_direction() );
    EXPECT_VECTOR_DOUBLE_EQ( read_back.coordinate_frame_v_direction(),
                             pinhole.coordinate_frame_v_direction() );
    EXPECT_VECTOR_DOUBLE_EQ( read_back.coordinate_frame_w_direction(),
                             pinhole.coordinate_frame_w_direction() );
    EXPECT_STREQ( read_back.lens_distortion()->name().c_str(),
                  pinhole.lens_distortion()->name().c_str() );
    EXPECT_VECTOR_DOUBLE_EQ( read_back.lens_distortion()->distortion_parameters(),
                             pinhole.lens_distortion()->distortion_parameters() );
#endif
  }

  PinholeModel pinhole;
  Matrix<double> expect_rot;
};

TEST_F( PinholeTest, NullLensDistortion ) {
  projection_test();
  UnlinkName file("NullCam.pinhole");
  readback_test( file );

  // check enforcement that pose returns the rotation from camera
  // frame to world frame.
  Vector2 center_pixel = pinhole.point_offset();
  Quaternion<double> center_pose =
    pinhole.camera_pose(center_pixel);
  double angle_from_z =
    acos(dot_prod(Vector3(0,0,1),inverse(center_pose).rotate(pinhole.pixel_to_vector(center_pixel))));
  EXPECT_LT( angle_from_z, 0.5 );
}

TEST_F( PinholeTest, TsaiLensDistortion ) {
  pinhole.set_lens_distortion(
    TsaiLensDistortion(Vector4(-0.2796604335308075,
                               0.1031486615538597,
                               -0.0007824968779459596,
                               0.0009675505571067333) ) );
#if defined(VW_HAVE_PKG_LAPACK) && VW_HAVE_PKG_LAPACK==1
  projection_test(1e-4);
#endif
  UnlinkName file("TsaiCam.pinhole");
  readback_test( file );
}

TEST_F( PinholeTest, BrownConradyDistortion ) {
  pinhole.set_lens_distortion(
    BrownConradyDistortion(Vector2(-0.6,-0.2),
                           Vector3(.1336185e-8,
                                   -0.5226175e-12,
                                   0),
                           Vector2(.5495819e-9,
                                   0),
                           0.201) );
#if defined(VW_HAVE_PKG_LAPACK) && VW_HAVE_PKG_LAPACK==1
  projection_test(1e-4);
#endif
  UnlinkName file("BrownConrady.pinhole");
  readback_test( file );
}

TEST_F( PinholeTest, AdjustableTsaiDistortion ) {
  Vector<double> distort_coeff(6);
  distort_coeff[0] = 0.007646500298509824;
  distort_coeff[1] = -0.01743067138801845;
  distort_coeff[2] = 0.00980946292640812;
  distort_coeff[3] = -2.98092556225311e-05;
  distort_coeff[4] = -1.339089765674149e-05;
  distort_coeff[5] = -1.221974557659228e-05;
  pinhole.set_lens_distortion(
     AdjustableTsaiLensDistortion( distort_coeff ) );
#if defined(VW_HAVE_PKG_LAPACK) && VW_HAVE_PKG_LAPACK==1
  projection_test(1e-4);
#endif
  UnlinkName file("AdjustedTsai.pinhole");
  readback_test( file );
}

TEST_F( PinholeTest, OldFormatReadTest ) {
  UnlinkName filename("monkey.tsai");
  std::ofstream filestream( filename.c_str() );
  filestream << "fu = 54.6\n";
  filestream << "fv = 45.3\n";
  filestream << "cu = 3\n";
  filestream << "cv = 5\n";
  filestream << "u_direction = 1 0 0\n";
  filestream << "v_direction = 0 1 0\n";
  filestream << "w_direction = 0 0 1\n";
  filestream << "C = 18.6 14.4 13.3\n";
  filestream << "R = 1 0 0 0 1 0 0 0 1\n";
  filestream << "k1 = 0.001\n";
  filestream << "k2 = 0.001\n";
  filestream << "p1 = 0.010\n";
  filestream << "p2 = 1\n";
  filestream.close();

  PinholeModel monkey;
  ASSERT_NO_THROW( monkey.read( filename ) );
  EXPECT_VECTOR_NEAR( Vector2(54.6,45.3),
                      monkey.focal_length(), 1e-5 );
  EXPECT_VECTOR_NEAR( Vector2(3,5),
                      monkey.point_offset(), 1e-5 );
  EXPECT_VECTOR_NEAR( Vector3(18.6,14.4,13.3),
                      monkey.camera_center(Vector2()), 1e-5 );
}
